//=======================================================================================
// ____          ____    __    ______     __________   __      __       __        __
// \    \       |    |  |  |  |   _   \  |___    ___| |  |    |  |     /  \      |  |
//  \    \      |    |  |  |  |  |_)   |     |  |     |  |    |  |    /    \     |  |
//   \    \     |    |  |  |  |   _   /      |  |     |  |    |  |   /  /\  \    |  |
//    \    \    |    |  |  |  |  | \  \      |  |     |   \__/   |  /  ____  \   |  |____
//     \    \   |    |  |__|  |__|  \__\     |__|      \________/  /__/    \__\  |_______|
//      \    \  |    |   ________________________________________________________________
//       \    \ |    |  |  ______________________________________________________________|
//        \    \|    |  |  |         __          __     __     __     ______      _______
//         \         |  |  |_____   |  |        |  |   |  |   |  |   |   _  \    /  _____)
//          \        |  |   _____|  |  |        |  |   |  |   |  |   |  | \  \   \_______
//           \       |  |  |        |  |_____   |   \_/   |   |  |   |  |_/  /    _____  |
//            \ _____|  |__|        |________|   \_______/    |__|   |______/    (_______/
//
//  This file is part of VirtualFluids. VirtualFluids is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  VirtualFluids is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with VirtualFluids (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \file AdvectionDiffusion.h
//! \ingroup AdvectionDiffusion
//! \author Martin Schoenherr
//=======================================================================================
#include "AdvectionDiffusion/AdvectionDiffusion.h"
#include "GPU/CudaMemoryManager.h"
#include "GPU/GPU_Interface.h"
#include "Parameter/Parameter.h"

////////////////////////////////////////////////////////////////////////////////
void initAD(SPtr<Parameter> para)
{
    //////////////////////////////////////////////////////////////////////////
    // calculation of omega for diffusivity
    para->getParD()->omegaD = (real)2.0 / ((real)6.0 * para->getParD()->diffusivity + (real)1.0);
    //////////////////////////////////////////////////////////////////////////
    para->getParD()->isEvenTimestep = true;
    //////////////////////////////////////////////////////////////////////////
    InitADdevice(
        para->getParD()->numberofthreads, 
        para->getParD()->neighborX, 
        para->getParD()->neighborY,
        para->getParD()->neighborZ, 
        para->getParD()->typeOfGridNode, 
        para->getParD()->concentration,
        para->getParD()->velocityX, 
        para->getParD()->velocityY, 
        para->getParD()->velocityZ,
        para->getParD()->numberOfNodes, 
        para->getParD()->distributionsAD.f[0],
        para->getParD()->isEvenTimestep);
    //////////////////////////////////////////////////////////////////////////
    para->getParD()->isEvenTimestep = false;
    //////////////////////////////////////////////////////////////////////////
    InitADdevice(
        para->getParD()->numberofthreads, 
        para->getParD()->neighborX, 
        para->getParD()->neighborY,
        para->getParD()->neighborZ, 
        para->getParD()->typeOfGridNode, 
        para->getParD()->concentration,
        para->getParD()->velocityX, 
        para->getParD()->velocityY, 
        para->getParD()->velocityZ,
        para->getParD()->numberOfNodes, 
        para->getParD()->distributionsAD.f[0],
        para->getParD()->isEvenTimestep);
    //////////////////////////////////////////////////////////////////////////
    CalcConcentration27(
        para->getParD()->numberofthreads,
        para->getParD()->concentration,
        para->getParD()->typeOfGridNode,
        para->getParD()->neighborX,
        para->getParD()->neighborY,
        para->getParD()->neighborZ,
        para->getParD()->numberOfNodes,
        para->getParD()->distributionsAD.f[0],
        para->getParD()->isEvenTimestep);
}

////////////////////////////////////////////////////////////////////////////////
void setInitialNodeValuesAD(SPtr<Parameter> para, SPtr<CudaMemoryManager> cudaMemoryManager)
{
    for (uint j = 1; j <= para->getParH()->numberOfNodes; j++) {
        const real coordX = para->getParH()->coordinateX[j];
        const real coordY = para->getParH()->coordinateY[j];
        const real coordZ = para->getParH()->coordinateZ[j];

        real concentration;

        // call functor object with initial condition
        if (para->getInitialConditionAD()) {
            para->getInitialConditionAD()(coordX, coordY, coordZ, concentration);
        } else {
            concentration = real(0.0);
        }

        para->getParH()->concentration[j] = concentration;
    }

    cudaMemoryManager->cudaCopyConcentrationHostToDevice();
}

////////////////////////////////////////////////////////////////////////////////
void calcAD(SPtr<Parameter> para)
{
    FactorizedCentralMomentsAdvectionDiffusionDeviceKernel(
        para->getParD()->numberofthreads,
        para->getParD()->omegaD,
        para->getParD()->typeOfGridNode,
        para->getParD()->neighborX,
        para->getParD()->neighborY,
        para->getParD()->neighborZ,
        para->getParD()->distributions.f[0],
        para->getParD()->distributionsAD.f[0],
        para->getParD()->numberOfNodes,
        para->getParD()->forcing,
        para->getParD()->isEvenTimestep);

    if (para->getParD()->numberOfSlipBCnodes > 1) {
        ADSlipVelDevComp(
            para->getParD()->numberofthreads,
            para->getParD()->slipBC.NormalX,
            para->getParD()->slipBC.NormalY,
            para->getParD()->slipBC.NormalZ,
            para->getParD()->distributions.f[0],
            para->getParD()->distributionsAD.f[0],
            para->getParD()->slipBC.k,
            para->getParD()->slipBC.q27[0],
            para->getParD()->numberOfSlipBCnodes,
            para->getParD()->omegaD,
            para->getParD()->neighborX,
            para->getParD()->neighborY,
            para->getParD()->neighborZ,
            para->getParD()->numberOfNodes,
            para->getParD()->isEvenTimestep);
    }
}

////////////////////////////////////////////////////////////////////////////////
void printAD(SPtr<Parameter> para, SPtr<CudaMemoryManager> cudaMemoryManager)
{
    CalcConcentration27(
        para->getParD()->numberofthreads,
        para->getParD()->concentration,
        para->getParD()->typeOfGridNode,
        para->getParD()->neighborX,
        para->getParD()->neighborY,
        para->getParD()->neighborZ,
        para->getParD()->numberOfNodes,
        para->getParD()->distributionsAD.f[0],
        para->getParD()->isEvenTimestep);

    cudaMemoryManager->cudaCopyConcentrationDeviceToHost();
}
